"""Filter reserved tags on reading an index.

The prefix 'pidx:' for tags is reserved for internal use in photoidx.
It should be removed when reading an index file.
"""

import filecmp
import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

invindex = gettestdata("index-reserved-tags.yaml")
refindex = gettestdata("index-tagged.yaml")

def test_reserved_tags_convert(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    idxfile = str(tmpdir / ".index.yaml")
    shutil.copy(invindex, idxfile)
    # reading and writing the index transparantly filters out tags
    # using the reserved prefix.
    with photoidx.index.Index(idxfile=tmpdir) as idx:
        idx.write()
    assert filecmp.cmp(idxfile, refindex), "index file differs from reference"



"""Create an image index.
"""

import filecmp
import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

refindex = gettestdata("index-create.yaml")

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    return tmpdir

def test_create_curdir(imgdir, monkeypatch):
    """Create a new index in the current directory adding all images.
    """
    monkeypatch.chdir(str(imgdir))
    with photoidx.index.Index(imgdir=".") as idx:
        idx.write()
    idxfile = ".index.yaml"
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"

@pytest.mark.dependency()
def test_create(imgdir):
    """Create a new index adding all images in the imgdir.
    """
    with photoidx.index.Index(imgdir=imgdir) as idx:
        idx.write()
    idxfile = str(imgdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"

@pytest.mark.dependency(depends=["test_create"])
def test_read(imgdir):
    """Read the index file and write it out again.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idx.write()
    idxfile = str(imgdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"



"""Provide the class Image corresponding to an IdxItem.
"""

import logging
import re
from PySide2 import QtCore, QtGui, QtWidgets
try:
    import vignette
except ImportError:
    vignette = None

log = logging.getLogger(__name__)

# Limit the vignette thumbnailer backends to those dealing with images.
if vignette:
    try:
        # vignette 4.5.2 or newer
        vignette.select_thumbnailer_types([vignette.FILETYPE_IMAGE])
    except AttributeError:
        if hasattr(vignette, 'THUMBNAILER_BACKENDS'):
            # vignette 4.3.0 or newer
            vignette.THUMBNAILER_BACKENDS = [
                vignette.QtBackend(),
                vignette.PilBackend(),
                vignette.MagickBackend()
            ]
        else:
            # vignette is too old to be usable
            vignette = None
    if vignette:
        if not list(vignette.iter_thumbnail_backends()):
            log.warning("Disabling vignette: "
                        "no suitable thumbnailer backend available")
            vignette = None


class ImageNotFoundError(Exception):
    pass


class Image(object):

    ThumbnailSize = QtCore.QSize(128, 128)

    def __init__(self, basedir, item):
        self.item = item
        self.fileName = basedir / item.filename
        self.name = item.name or self.fileName.name
        self.transform = QtGui.QMatrix()
        if self.item.orientation:
            m = re.match(r"Rotate (\d+) CW", self.item.orientation)
            if m:
                self.rotate(int(m.group(1)))

    def getPixmap(self):
        image = QtGui.QImage(str(self.fileName))
        if image.isNull():
            raise ImageNotFoundError("Cannot load %s." % self.fileName)
        return QtGui.QPixmap.fromImage(image).transformed(self.transform)

    def getThumbPixmap(self):
        if vignette:
            thumbpath = vignette.get_thumbnail(str(self.fileName), 'normal')
            image = QtGui.QImage(thumbpath)
            if image.isNull():
                raise ImageNotFoundError("Cannot load %s." % self.fileName)
            pixmap = QtGui.QPixmap.fromImage(image)
        else:
            image = QtGui.QImage(str(self.fileName))
            if image.isNull():
                raise ImageNotFoundError("Cannot load %s." % self.fileName)
            pixmap = QtGui.QPixmap.fromImage(image)
            pixmap = pixmap.scaled(self.ThumbnailSize, 
                                   QtCore.Qt.KeepAspectRatio, 
                                   QtCore.Qt.SmoothTransformation)
        pixmap = pixmap.transformed(self.transform)
        return pixmap

    def rotate(self, a):
        self.transform.rotate(a)



"""Test different options to filter images by date.
"""

import datetime
import pytest
import photoidx.index
import photoidx.idxfilter
from conftest import gettestdata

testimgs = [ "dsc_%04d.jpg" % i for i in range(1,13) ]
indexfile = gettestdata("index-date.yaml")


def test_single_date():
    """Select by single date.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        date = (datetime.datetime(2016, 2, 29), datetime.datetime(2016, 3, 1))
        idxfilter = photoidx.idxfilter.IdxFilter(date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[1:4]


def test_interval_date_date():
    """Select by an interval between two dates.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        date = (datetime.datetime(2016, 2, 29), datetime.datetime(2016, 3, 6))
        idxfilter = photoidx.idxfilter.IdxFilter(date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[1:11]


def test_interval_date_datetime():
    """Select by an interval between start date and end date/time.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        date = (datetime.datetime(2016, 2, 29), 
                datetime.datetime(2016, 3, 5, 3, 47, 9))
        idxfilter = photoidx.idxfilter.IdxFilter(date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[1:9]


def test_single_datetime():
    """Select by single date/time.

    Probably not very useful in the praxis, but valid.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        date = (datetime.datetime(2016, 3, 3, 11, 21, 40), 
                datetime.datetime(2016, 3, 3, 11, 21, 41))
        idxfilter = photoidx.idxfilter.IdxFilter(date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[6:7]


def test_interval_datetime_date():
    """Select by an interval between start date/time and end date.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        date = (datetime.datetime(2016, 3, 3, 11, 21, 40), 
                datetime.datetime(2016, 3, 6))
        idxfilter = photoidx.idxfilter.IdxFilter(date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[6:11]


def test_interval_datetime_datetime():
    """Select by an interval between two date/times.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        date = (datetime.datetime(2016, 3, 3, 11, 21, 41), 
                datetime.datetime(2016, 3, 5, 3, 47, 9))
        idxfilter = photoidx.idxfilter.IdxFilter(date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[7:9]



"""Add images to an index.
"""

import filecmp
import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

refindex = gettestdata("index-create.yaml")

def test_createupdate(tmpdir):
    for fname in testimgfiles[:3]:
        shutil.copy(fname, str(tmpdir))
    with photoidx.index.Index(imgdir=tmpdir) as idx:
        idx.write()
    for fname in testimgfiles[3:]:
        shutil.copy(fname, str(tmpdir))
    with photoidx.index.Index(idxfile=tmpdir, imgdir=tmpdir) as idx:
        idx.write()
    idxfile = str(tmpdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"



"""Read an image index from the index file and write it back.
"""

import filecmp
import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

refindex = gettestdata("index-tagged.yaml")
refindexu = gettestdata("index-unicode-tags.yaml")

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    return tmpdir

def test_read_non_existent(imgdir):
    """Try to read an index file that does not exist.
    """
    with pytest.raises(OSError):
        with photoidx.index.Index(idxfile=imgdir) as idx:
            pass

def test_read_write(imgdir):
    """Read the index file and write it out again.
    """
    idxfile = str(imgdir / ".index.yaml")
    shutil.copy(refindex, idxfile)
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idx.write()
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"

def test_read_write_unicode(imgdir):
    """Same test as above but with non-ASCII characters in the tags.
    """
    idxfile = str(imgdir / ".index.yaml")
    shutil.copy(refindexu, idxfile)
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idx.write()
    assert filecmp.cmp(refindexu, idxfile), "index file differs from reference"



"""Concurrent access to the index file.

The Index class is supposed to protect the index file against
conflicts in concurrent file access using file system locking.
We will need multiple processes to test this.
"""

import filecmp
from multiprocessing import Process, Queue
import shutil
import pytest
import photoidx.index
import photoidx.idxfilter
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]
refindex = gettestdata("index-tagged.yaml")

# tags to test and expected results.
tags = {
    "Shinto_shrine": ["dsc_4664.jpg", "dsc_4831.jpg"],
    "Tokyo": ["dsc_4623.jpg", "dsc_4664.jpg"],
}

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    shutil.copy(refindex, str(tmpdir / ".index.yaml"))
    return tmpdir

# --------------------------------------------------------------------
# Worker functions to be run in the subprocesses.  It does not really
# matter what we do here.  We just want two examples doing some
# readonly and some write access to the index file respectively.
#
# We use queues to synchronize in order to make sure the concurrent
# invocations access the index really at the same moment.

def ls_bytag(imgdir, tag, qres, qwait):
    """List files by tag.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(tags=tag)
        qres.put((tag, [ str(i.filename) for i in idxfilter.filter(idx) ]))
        qwait.get()

def add_tag(imgdir, tag, qres):
    """Add a tag to all items.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        for i in idx:
            i.tags.add(tag)
        try:
            idx.write()
        except Exception as e:
            qres.put(e)
        else:
            qres.put("Done")

# --------------------------------------------------------------------

def test_concurrent_read(imgdir):
    """Test concurrent read access.
    Should work without problem.
    """
    qres = Queue()
    qwait = Queue()
    procs = []
    for t in tags.keys():
        p = Process(target=ls_bytag, args=(imgdir, t, qres, qwait))
        p.start()
        procs.append(p)
    print("Reading processes started.")
    for i in range(len(procs)):
        (t, files) = qres.get()
        assert files == tags[t]
    print("Replys from reading processes received.")
    # At this point, all subprocesses have the index file open for
    # reading: they sent a reply, so they have opened it, but they
    # still wait for the message on qwait, so they have not yet closed
    # it.
    for i in range(len(procs)):
        qwait.put("done")
    print("Joining processes.")
    for p in procs:
        p.join()

def test_concurrent_read_write(imgdir):
    """Test writing the index while another process is reading it.
    This should fail.
    """
    qresr = Queue()
    qresw = Queue()
    qwait = Queue()
    procs = []
    p = Process(target=ls_bytag, args=(imgdir, "Tokyo", qresr, qwait))
    p.start()
    procs.append(p)
    print("Reading process started.")
    (t, files) = qresr.get()
    assert files == tags[t]
    print("Reply from reading process received.")
    # At this point, the first subprocesse has the index file open for
    # reading and waits for the message on qwait.  Start the second
    # one that tries to write it.
    p = Process(target=add_tag, args=(imgdir, "all", qresw))
    p.start()
    procs.append(p)
    print("Writing process started.")
    # Verify that the writing process caught an AlreadyLockedError.
    r = qresw.get()
    assert isinstance(r, photoidx.index.AlreadyLockedError)
    print("Reply from writing process received.")
    # Now, allow the reading process to close the index file.
    qwait.put("done")
    print("Joining processes.")
    for p in procs:
        p.join()
    # Finally, verify that the index is still unchanged, so the
    # writing did not succeed.
    idxfile = str(imgdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"



"""Check checksums.
"""

from pathlib import Path
import shutil
import subprocess
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

hashalg = {
    "md5": "/usr/bin/md5sum",
    "sha1": "/usr/bin/sha1sum",
    "sha224": "/usr/bin/sha224sum",
    "sha256": "/usr/bin/sha256sum",
    "sha384": "/usr/bin/sha384sum",
    "sha512": "/usr/bin/sha512sum",
}

@pytest.mark.dependency()
def test_create_checksum(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    with photoidx.index.Index(imgdir=tmpdir, hashalg=hashalg.keys()) as idx:
        idx.write()

@pytest.mark.dependency(depends=["test_create_checksum"])
@pytest.mark.parametrize("alg", hashalg.keys())
def test_check_checksum(tmpdir, monkeypatch, alg):
    checkprog = hashalg[alg]
    if not Path(checkprog).is_file():
        pytest.skip("%s not found." % checkprog)
    fname = tmpdir / alg
    with photoidx.index.Index(idxfile=tmpdir) as idx, fname.open("wt") as f:
        for i in idx:
            print("%s  %s" % (i.checksum[alg], i.filename), file=f)
    monkeypatch.chdir(str(tmpdir))
    with fname.open("rt") as f:
        cmd = [checkprog, "-c"]
        print(">", *cmd)
        subprocess.check_call(cmd, stdin=f)

def test_no_checksum(tmpdir):
    with photoidx.index.Index(imgdir=tmpdir, hashalg=[]) as idx:
        idx.write()
    with photoidx.index.Index(idxfile=tmpdir) as idx:
        for i in idx:
            assert i.checksum == {}



"""Read legacy format index files.

Since version 0.4 different checksum algorithms are supported (Issue
#12).  As a consequence, the index file format is changed.  photoidx
still reads legacy files and transparently converts them to the new
format.  This feature is tested in this module.
"""

import filecmp
import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

legacyindex = gettestdata("index-legacy.yaml")
refindex = gettestdata("index-create.yaml")

def test_legacyconvert(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    idxfile = str(tmpdir / ".index.yaml")
    shutil.copy(legacyindex, idxfile)
    # reading and writing the index transparantly converts it.
    with photoidx.index.Index(idxfile=tmpdir) as idx:
        idx.write()
    assert filecmp.cmp(idxfile, refindex), "index file differs from reference"



"""Provide the class IdxItem which represents an item in the index.
"""

import hashlib
from pathlib import Path
from photoidx.exif import Exif
from photoidx.geo import GeoPosition


def _checksum(fname, hashalg):
    """Calculate hashes for a file.
    """
    if not hashalg:
        return {}
    m = { h:hashlib.new(h) for h in hashalg }
    chunksize = 8192
    with fname.open('rb') as f:
        while True:
            chunk = f.read(chunksize)
            if not chunk:
                break
            for h in hashalg:
                m[h].update(chunk)
    return { h: m[h].hexdigest() for h in hashalg }


class IdxItem(object):

    def __init__(self, data=None, filename=None, basedir=None, hashalg=['md5']):
        if data is not None:
            self.filename = Path(data.get('filename'))
            self.name = data.get('name', None)
            self.checksum = data.get('checksum', {})
            if not self.checksum and 'md5' in data:
                # legacy: old index file format used to have a 'md5'
                # attribute, rather then 'checksum'.
                self.checksum['md5'] = data['md5']
            self.createDate = data.get('createDate')
            if self.createDate is None and 'createdate' in data:
                # legacy: 'createDate' used to be 'createdate' in old
                # index file format.
                self.createDate = data['createdate']
            self.orientation = data.get('orientation')
            self.gpsPosition = data.get('gpsPosition')
            tags = data.get('tags', [])
            self.tags = set(filter(lambda t: not t.startswith('pidx:'), tags))
            self.selected = 'pidx:selected' in tags
        elif filename is not None:
            filename = Path(filename)
            self.filename = filename
            self.name = None
            if basedir is not None:
                filename = Path(basedir) / filename
            self.checksum = _checksum(filename, hashalg)
            exifdata = Exif(filename)
            self.createDate = exifdata.createDate
            self.orientation = exifdata.orientation
            self.gpsPosition = exifdata.gpsPosition
            self.tags = set()
            self.selected = False
        if self.gpsPosition:
            self.gpsPosition = GeoPosition(self.gpsPosition)

    def as_dict(self):
        tags = self.tags.copy()
        if self.selected:
             tags.add('pidx:selected')
        d = {
            'filename': str(self.filename),
            'checksum': self.checksum,
            'createDate': self.createDate,
            'orientation': self.orientation,
            'gpsPosition': self.gpsPosition,
            'tags': sorted(tags),
        }
        if d['gpsPosition']:
            d['gpsPosition'] = d['gpsPosition'].as_dict()
        if self.name is not None:
            d['name'] = self.name
        return d



"""Test different options to filter images by date.

This is essentially the same as test_01_filter_date, but using
IdxFilter.from_args() rather then the constructor directly.  So it
mostly test the date parser in idxfilter.
"""

import argparse
import pytest
import photoidx.index
import photoidx.idxfilter
from conftest import gettestdata

testimgs = [ "dsc_%04d.jpg" % i for i in range(1,13) ]
indexfile = gettestdata("index-date.yaml")

@pytest.fixture(scope="module")
def argparser():
    parser = argparse.ArgumentParser()
    photoidx.idxfilter.addFilterArguments(parser)
    return parser


def test_single_date(argparser):
    """Select by single date.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        args = argparser.parse_args(["--date=2016-02-29"])
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[1:4]


def test_interval_date_date(argparser):
    """Select by an interval between two dates.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        args = argparser.parse_args(["--date=2016-02-29--2016-03-06"])
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[1:11]


def test_interval_date_datetime(argparser):
    """Select by an interval between start date and end date/time.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        args = argparser.parse_args(["--date=2016-02-29/2016-03-05T03:47:09"])
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[1:9]


def test_single_datetime(argparser):
    """Select by single date/time.

    Probably not very useful in the praxis, but valid.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        args = argparser.parse_args(["--date=2016-03-03T11:21:40"])
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[6:7]


def test_interval_datetime_date(argparser):
    """Select by an interval between start date/time and end date.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        args = argparser.parse_args(["--date=2016-03-03T11:21:40/2016-03-06"])
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[6:11]


def test_interval_datetime_datetime(argparser):
    """Select by an interval between two date/times.
    """
    with photoidx.index.Index(idxfile=indexfile) as idx:
        args = argparser.parse_args(["--date=2016-03-03T11:21:41"
                                     "--2016-03-05T03:47:09"])
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == testimgs[7:9]



"""Wrapper around the exif library to extract and convert some information.
"""

import datetime
import fractions
import warnings
import exifread

# Some helper classes for exif attributes, having customized string
# representations.

class ExposureTime(fractions.Fraction):
    def __str__(self):
        if self.denominator == 1:
            return "%s sec" % (self.numerator)
        else:
            return "%s/%s sec" % (self.numerator, self.denominator)

class Aperture(float):
    def __str__(self):
        # In fact, we display the reciprocal of the aperture, the
        # f-number, in the string representation.
        if self == int(self):
            return "f/%d" % (int(self))
        else:
            return "f/%.1f" % (float(self))

class FocalLength(float):
    def __str__(self):
        if self == int(self):
            return "%d mm" % (int(self))
        else:
            return "%.1f mm" % (float(self))


class Exif(object):

    OrientationXlate = {
        1: 'Horizontal (normal)',
        2: 'Mirror horizontal',
        3: 'Rotate 180',
        4: 'Mirror vertical',
        5: 'Mirror horizontal and rotate 270 CW',
        6: 'Rotate 90 CW',
        7: 'Mirror horizontal and rotate 90 CW',
        8: 'Rotate 270 CW',
    }

    def __init__(self, path):
        with path.open("rb") as f:
            self._tags = exifread.process_file(f)

    @property
    def createDate(self):
        """Time and date the image was taken."""
        try:
            dt = self._tags['Image DateTime'].values
        except (AttributeError, KeyError):
            return None
        else:
            return datetime.datetime.strptime(dt, "%Y:%m:%d %H:%M:%S")

    @property
    def orientation(self):
        """Orientation of the camera relative to the scene."""
        try:
            orientation = self._tags['Image Orientation'].values[0]
        except (AttributeError, KeyError):
            return None
        else:
            return self.OrientationXlate[int(orientation)]

    @property
    def gpsPosition(self):
        """GPS coordinates."""
        try:
            lat_tuple = self._tags['GPS GPSLatitude'].values
            lon_tuple = self._tags['GPS GPSLongitude'].values
            latref = self._tags['GPS GPSLatitudeRef'].values
            lonref = self._tags['GPS GPSLongitudeRef'].values
        except (AttributeError, KeyError):
            return None
        else:
            lat = lat_tuple[0] + lat_tuple[1]/60 + lat_tuple[2]/3600
            lon = lon_tuple[0] + lon_tuple[1]/60 + lon_tuple[2]/3600
            return { latref:float(lat), lonref:float(lon) }

    @property
    def cameraModel(self):
        """Camera Model."""
        try:
            return self._tags['Image Model'].values
        except (AttributeError, KeyError):
            return None

    @property
    def exposureTime(self):
        """Exposure time."""
        try:
            et = self._tags['EXIF ExposureTime'].values[0]
        except (AttributeError, KeyError):
            return None
        else:
            return ExposureTime(et)

    @property
    def aperture(self):
        """Aperture."""
        try:
            f = self._tags['EXIF FNumber'].values[0]
        except (AttributeError, KeyError):
            return None
        else:
            return Aperture(f)

    @property
    def iso(self):
        """ISO speed rating."""
        try:
            return self._tags['EXIF ISOSpeedRatings'].values[0]
        except (AttributeError, KeyError):
            return None

    @property
    def focalLength(self):
        """Lens focal length."""
        try:
            fl = self._tags['EXIF FocalLength'].values[0]
        except (AttributeError, KeyError):
            return None
        else:
            return FocalLength(fl)



"""Provide the class Stats that collects some statistics.
"""

import datetime


class Stats(object):

    def __init__(self, items):
        self.count = 0
        self.selected = 0
        self.oldest = datetime.datetime.max
        self.newest = datetime.datetime.min
        self.by_date = {}
        self.by_tag = {}
        for i in items:
            self.count += 1
            if i.selected:
                self.selected += 1
            if i.createDate:
                if i.createDate < self.oldest:
                    self.oldest = i.createDate
                if i.createDate > self.newest:
                    self.newest = i.createDate
                date = i.createDate.toordinal()
                self.by_date.setdefault(date, 0)
                self.by_date[date] += 1
            for tag in i.tags:
                self.by_tag.setdefault(tag, 0)
                self.by_tag[tag] += 1

    def __bool__(self):
        return bool(self.count)

    def __str__(self):
        s = "Count: %d\nSelected: %d\n" % (self.count, self.selected)
        if self.newest >= self.oldest:
            s += "Oldest: %s\n" % str(self.oldest)
            s += "Newest: %s\n" % str(self.newest)
        if self.by_date:
            s += "By date:\n"
            for d in sorted(self.by_date.keys()):
                s += ("  %s: %d\n" 
                      % (datetime.date.fromordinal(d), self.by_date[d]))
        if self.by_tag:
            s += "By tag:\n"
            for t in sorted(self.by_tag.keys()):
                s += ("  %s: %d\n" % (t, self.by_tag[t]))
        return s



"""Some useful list classes.

**Note**: This module might be useful independently of photoidx.  It
is included here because photoidx uses it internally, but it is not
considered to be part of the API.  Changes in this module are not
considered API changes of photoidx.  It may even be removed from
future versions of the photoidx distribution without further notice.
"""

from collections.abc import MutableSequence

class LazyList(MutableSequence):
    """A list generated lazily from an iterable.

    LazyList provides list access to the sequence of elements from the
    iterable.  Elements are taken out lazily.  That means, the
    elements are taken from the iterable not before they are actually
    accessed.  Once taken out, the elements are stored in a
    conventional list in order to provide random access.  The string
    representation operator of LazyList only displays the elements
    taken out of the iterable so far.

    Note: if the list is accessed at the end using negative indices,
    all elements are taken from the iterable before returning the
    result.  Some operations implicitly access the list at the end and
    thus take all elements from the iterable.  These operations
    include `len()` and `append()`.  Do not access the list at the end
    using negativ indices or append to the list if you cannot afford
    to take all elements out of the iterable.

    >>> l = LazyList((0, 1, 2, 3, 4))
    >>> l
    []
    >>> l[1]
    1
    >>> l
    [0, 1]
    >>> del l[1]
    >>> l
    [0]
    >>> l[1]
    2
    >>> l
    [0, 2]
    >>> l[8]
    Traceback (most recent call last):
      ...
    IndexError: list index out of range
    >>> l
    [0, 2, 3, 4]
    >>> l = LazyList((0, 1, 2, 3, 4))
    >>> l[-2]
    3
    >>> l
    [0, 1, 2, 3, 4]
    >>> l = LazyList((0, 1, 2, 3, 4))
    >>> list(l) == [0, 1, 2, 3, 4]
    True
    >>> l
    [0, 1, 2, 3, 4]
    >>> l = LazyList((0, 1, 2, 3, 4))
    >>> len(l)
    5
    >>> l
    [0, 1, 2, 3, 4]
    >>> l = LazyList((0, 1, 2, 3, 4))
    >>> l.append(5)
    >>> l
    [0, 1, 2, 3, 4, 5]
    >>> def naturals():
    ...     n = 0
    ...     while True:
    ...         yield n
    ...         n += 1
    ... 
    >>> l = LazyList(naturals())
    >>> l[1]
    1
    >>> l
    [0, 1]
    >>> l[4:2:-1]
    [4, 3]
    >>> l
    [0, 1, 2, 3, 4]
    >>> l[8]
    8
    >>> l
    [0, 1, 2, 3, 4, 5, 6, 7, 8]
    >>> l[17:11]
    []
    >>> l
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    >>> l = LazyList(naturals())
    >>> l
    []
    >>> bool(l)
    True
    >>> l
    [0]
    >>> l = LazyList((0, 1, 2, 3, 4, 5, 6))
    >>> l.index(1)
    1
    >>> l
    [0, 1]
    >>> l.index(2, 1)
    2
    >>> l
    [0, 1, 2]
    >>> l.index(2, 3, 5)
    Traceback (most recent call last):
      ...
    ValueError: 2 is not in list
    >>> l
    [0, 1, 2, 3, 4]
    >>> l.index(2, 3)
    Traceback (most recent call last):
      ...
    ValueError: 2 is not in list
    >>> l
    [0, 1, 2, 3, 4, 5, 6]

    """

    def __init__(self, iterable):
        self.iterable = iter(iterable)
        self.elements = []

    def _access_next(self):
        """Take out the next element from the iterable.
        Raise StopIteration if the iterable is exhausted.
        """
        v = next(self.iterable)
        self.elements.append(v)
        return v

    def _access(self, index):
        """Try to take out the elements covered by index from the iterable.
        The argument may be an int or a slice.  Do not raise an error,
        even if not enough elements can be delivered by the iterable.
        """
        m = 0
        if isinstance(index, int):
            m = index + 1 if index >= 0 else -1
        elif isinstance(index, slice):
            if index.step is not None and index.step < 0:
                m = index.start + 1 if index.start >= 0 else -1
            else:
                m = index.stop if index.stop >= 0 else -1
        while len(self.elements) < m or m < 0:
            try:
                self._access_next()
            except StopIteration:
                break

    def __len__(self):
        self._access(-1)
        return len(self.elements)

    def __getitem__(self, index):
        self._access(index)
        return self.elements.__getitem__(index)

    def __setitem__(self, index, value):
        self._access(index)
        self.elements.__setitem__(index, value)

    def __delitem__(self, index):
        self._access(index)
        self.elements.__delitem__(index)

    def index(self, value, *args):
        try:
            startidx = args[0]
        except IndexError:
            startidx = 0
        try:
            endidx = args[1]
        except IndexError:
            endidx = None
        self._access(startidx)
        try:
            return self.elements.index(value, *args)
        except ValueError:
            i = len(self.elements)
            while True:
                if endidx and i >= endidx:
                    raise ValueError("%s is not in list" % repr(value))
                try:
                    v = self._access_next()
                    if v == value:
                        return i
                    i += 1
                except StopIteration:
                    raise ValueError("%s is not in list" % repr(value))

    def insert(self, index, value):
        self._access(index)
        self.elements.insert(index, value)

    def append(self, value):
        self._access(-1)
        self.elements.append(value)

    def __bool__(self):
        self._access(0)
        return bool(self.elements)

    def __str__(self):
        return str(self.elements)

    def __repr__(self):
        return repr(self.elements)



"""A dialog window to set and remove tags.
"""

import math
from PySide2 import QtCore, QtWidgets


class TagSelectDialog(QtWidgets.QDialog):

    def __init__(self, taglist):
        super().__init__()

        self.checkLayout = QtWidgets.QGridLayout()
        self.settags(taglist)

        self.entry = QtWidgets.QLineEdit()
        self.entry.returnPressed.connect(self.newtag)
        entryLabel = QtWidgets.QLabel("New tag:")
        entryLabel.setBuddy(self.entry)
        entryLayout = QtWidgets.QHBoxLayout()
        entryLayout.addWidget(entryLabel)
        entryLayout.addWidget(self.entry)

        btn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        buttonBox = QtWidgets.QDialogButtonBox(btn)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        mainLayout = QtWidgets.QVBoxLayout()
        mainLayout.addLayout(self.checkLayout)
        mainLayout.addLayout(entryLayout)
        mainLayout.addWidget(buttonBox, alignment=QtCore.Qt.AlignHCenter)
        self.setLayout(mainLayout)
        self.setWindowTitle("Select tags")

    def settags(self, tags):
        self.taglist = set(tags)
        self.tagCheck = {}
        # We need to rearrange the checkbox widgets in
        # self.checkLayout.  To this end, remove all old checkboxes
        # first, store those to be retained to self.tagCheck and then
        # add the new set of checkboxes to self.checkLayout, reusing
        # those in self.tagCheck.
        while True:
            child = self.checkLayout.takeAt(0)
            if not child:
                break
            cb = child.widget()
            t = cb.text()
            if t in self.taglist:
                self.tagCheck[t] = cb
        if len(self.taglist) > 0:
            ncol = int(math.ceil(math.sqrt(len(self.taglist)/10)))
            nrow = int(math.ceil(len(self.taglist)/ncol))
            c = 0
            for t in sorted(self.taglist):
                if t not in self.tagCheck:
                    self.tagCheck[t] = QtWidgets.QCheckBox(t)
                cb = self.tagCheck[t]
                self.checkLayout.addWidget(cb, c % nrow, c // nrow)
                c += 1

    def newtag(self):
        t = self.entry.text()
        if t not in self.taglist:
            self.taglist.add(t)
            self.settags(self.taglist)
        self.tagCheck[t].setCheckState(QtCore.Qt.Checked)
        self.adjustSize()
        self.entry.setText("")

    def setCheckedTags(self, tags):
        for t in self.taglist:
            state = QtCore.Qt.Checked if t in tags else QtCore.Qt.Unchecked
            self.tagCheck[t].setCheckState(state)

    def checkedTags(self):
        return { t for t in self.taglist if self.tagCheck[t].isChecked() }



"""Geo tools, in particular class GeoPosition.
"""

import re
import math
from collections.abc import Mapping


_geopos_pattern = (r"^\s*(?P<lat>\d+(?:\.\d*))\s*(?P<latref>N|S),\s*"
                   r"(?P<lon>\d+(?:\.\d*))\s*(?P<lonref>E|W)\s*$")
_geopos_re = re.compile(_geopos_pattern)


class deg(float):
    """A degree as float that can be converted to (degree, minute, second).
    """
    def __abs__(self):
        return deg(super().__abs__())
    def dms(self):
        dg = int(self)
        dm = int(60*(self-dg))
        ds = 60*(60*(self-dg)-dm)
        return (dg, dm, ds)

class lat(deg):
    """Latitude, which is either N or S.
    """
    def ref(self):
        return 'N' if self >= 0 else 'S'
    def dmsref(self):
        return abs(self).dms() + (self.ref(),)

class lon(deg):
    """Longitude, which is either E or W.
    """
    def ref(self):
        return 'E' if self >= 0 else 'W'
    def dmsref(self):
        return abs(self).dms() + (self.ref(),)


class GeoPosition(object):

    earthRadius = 6371.0
    """approximate radius of the earth in km."""

    def __init__(self, pos):
        if isinstance(pos, Mapping):
            if 'N' in pos and 'S' not in pos:
                self.lat = lat(pos['N'])
            elif 'S' in pos and 'N' not in pos:
                self.lat = lat(-pos['S'])
            else:
                raise ValueError("invalid Geo position %s: "
                                 "must have either 'N' or 'S' in keys." 
                                 % str(pos))
            if 'E' in pos and 'W' not in pos:
                self.lon = lon(pos['E'])
            elif 'W' in pos and 'E' not in pos:
                self.lon = lon(-pos['W'])
            else:
                raise ValueError("invalid Geo position %s: "
                                 "must have either 'E' or 'W' in keys." 
                                 % str(pos))
        elif isinstance(pos, str):
            m = _geopos_re.match(pos)
            if m:
                latsign = 1.0 if m.group('latref') == 'N' else -1.0
                self.lat = lat(latsign*float(m.group('lat')))
                lonsign = 1.0 if m.group('lonref') == 'E' else -1.0
                self.lon = lon(lonsign*float(m.group('lon')))
            else:
                raise ValueError("invalid Geo position '%s'." % pos)
        else:
            raise TypeError("invalid type '%s'" % type(pos))

    def __str__(self):
        return ("%d\xb0 %d\u2032 %.2f\u2033 %s, "
                "%d\xb0 %d\u2032 %.2f\u2033 %s"
                % (self.lat.dmsref() + self.lon.dmsref()))

    def floatstr(self):
        return "%.5f %s, %.5f %s" % (abs(self.lat), self.lat.ref(), 
                                     abs(self.lon), self.lon.ref())

    def as_dict(self):
        return {
            self.lat.ref(): float(abs(self.lat)),
            self.lon.ref(): float(abs(self.lon))
        }

    def as_osmurl(self, zoom=16):
        """Return an URL to OpenStreetMap displaying this position."""
        template = "http://www.openstreetmap.org/?mlat=%f&mlon=%f&zoom=%d"
        return template % (self.lat, self.lon, zoom)

    def __sub__(self, other):
        """Difference between two GeoPositions.

        Approximate distance of the two points on the earth surface in km, 
        simplifying assumed the earth to be a sphere.
        """
        if isinstance(other, GeoPosition):
            lat1 = math.pi * self.lat / 180.0
            lat2 = math.pi * other.lat / 180.0
            lon1 = math.pi * self.lon / 180.0
            lon2 = math.pi * other.lon / 180.0
            slat = math.sin((lat1-lat2)/2.0)
            slon = math.sin((lon1-lon2)/2.0)
            c1 = math.cos(lat1)
            c2 = math.cos(lat2)
            sigma = 2.0*math.asin(math.sqrt(slat*slat + c1*c2*slon*slon))
            return self.earthRadius * sigma
        else:
            return NotImplemented



"""Handle missing exif tags gracefully.  (Issue #15)

When creating a new index, the constructor of class IdxItem reads some
exif tags like DateTimeOriginal and Orientaion from the images to add
the information to the index.  But it should not raise an error if
some exif tags are missing.
"""

import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_1190.jpg", "dsc_4623.jpg", "dsc_4664.jpg", 
    "dsc_4831.jpg", "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    return tmpdir

def test_create(imgdir):
    with photoidx.index.Index(imgdir=imgdir) as idx:
        idx.write()
